/**         @file  memtrack.h
 *         @brief  Memory block metadata tracking headers
 *          @date  25/09/2006
 *        @author  François Cerbelle (Fanfan), francois@cerbelle.net
 *     @copyright  Copyright (c) 1997-2024, François Cerbelle
 *
 *  Originally inspired by "L'art du code", Steve Maguire, Microsoft Press
 *
 *   @internal
 *       Compiler  gcc
 *  Last modified  2024-07-28 20:46
 *   Organization  Cerbelle.net
 *        Company  Home
 *
 *   This source code is released for free distribution under the terms of the
 *   GNU General Public License as published by the Free Software Foundation.
 */

#ifndef __MEMTRACK_H__
#define __MEMTRACK_H__

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdlib.h>             /* size_t */
#include <stdint.h>             /* uint64_t */

#ifdef __cplusplus
extern "C" {
#endif

/** Memory block metadata list item
 *
 * Double linked list item to store memory block metadata
 */
typedef struct MemBlock {
    struct MemBlock *Prev;  /**< Previous item pointer */
    struct MemBlock *Next;  /**< Next item pointer */
    void *Ptr;              /**< Allocated memory block pointer */
    size_t Size;            /**< Allocated memory block size */
    char *File;             /**< Source file which asked the allocation */
    int Line;               /**< Source line number ch asked the allocation */
    char *CompilDate;       /**< Source file compilation date */
    char *CompilTime;       /**< Source file compilation time */
    char *Function;         /**< Fonction name which asked the allocation */
} TMemBlock;

/** Functor to register an allocated memory block metadata
 *
 * Create and adds a memory block metadata record in the tracking system to
 * detect memory leaks. It performs some basic sanity checks. The filename,
 * compilation date and compilation time can not be null or empty, the line
 * number can not be 0, the memory pointer to store can not be NULL or already
 * registered and its size needs to be greater than 0.
 *
 * The function is called from memdbg.c functions used in the application code
 * thru memory.h macros. The macro automatically fills the filename, line
 * number, compilation date and time, and function name (using GCC's non-ansi
 * __func_ extension).
 *
 * This functor is used to implement a lazy initialization. It initialy
 * reference a temporary function to trigger memtrack initialization before
 * calling the actual function. Then, it references the actual function directly
 * to avoid any useless tests.
 *
 *  @param [in] p_Ptr Allocated memory block pointer
 *  @param [in] p_Size Allocated memory block size
 *  @param [in] p_File : Source file
 *  @param [in] p_Line : Source line number
 *  @param [in] p_CompilDate : File compilation date
 *  @param [in] p_CompilTime : File compilation time
 *  @param [in] p_Function : Source function name
 *
 *  @return Registration status
 *      @retval 0 if succeeded,
 *      @retval 1 if not possible.
 */
extern unsigned int (*memtrack_addblock) (const void *p_Ptr,
        const size_t p_Size,
        const char *p_File,
        const int p_Line,
        const char *p_CompilDate,
        const char *p_CompilTime,
        const char *p_Function);

/** Functor to unregister an allocated memory block metadata
 *
 * Find and delete a memory block metadata record in the tracking system.
 * It performs some basic sanity checks. The filename, compilation date and
 * compilation time can not be null or empty, the line number can not be 0, the
 * memory pointer to remove can not be NULL, it needs to already be registered.
 *
 * The function is called from memdbg.c functions used in the application code
 * thru memory.h macros. The macro automatically fills the filename, line
 * number, compilation date and time, and function name (using GCC's non-ansi
 * __func_ extension).
 *
 * This functor is used to implement a lazy initialization. It initialy
 * reference a temporary function to trigger memtrack initialization before
 * calling the actual function. Then, it references the actual function directly
 * to avoid any useless tests.
 *
 *  @param [in] p_Ptr Allocated memory block pointer
 *  @param [in] p_File : Source file
 *  @param [in] p_Line : Source line number
 *  @param [in] p_CompilDate : File compilation date
 *  @param [in] p_CompilTime : File compilation time
 *  @param [in] p_Function : Source function name
 *
 *  @return Registration status
 *      @retval 0 if succeeded,
 *      @retval !0 if not possible.
 */
extern unsigned int (*memtrack_delblock) (const void *p_Ptr,
        const char *p_File,
        const int p_Line,
        const char *p_CompilDate,
        const char *p_CompilTime,
        const char *p_Function);

/** Functor to list allocated memory blocks metadata
 *
 * Dumps all the metadata of the registered memory blocks to stderr.
 *
 * This functor is used to implement a lazy initialization. It initialy
 * reference a temporary function to trigger memtrack initialization before
 * calling the actual function. Then, it references the actual function directly
 * to avoid any useless tests.
 *
 *  @return Number of registered blocks (counted)
 *      @retval 0 if succeeded,
 *      @retval !0 if not possible.
 */
extern uint64_t (*memtrack_dumpblocks) ();

/** Functor to get the number of allocated blocks
 *
 * This function returns the value of the internal counter of allocated memory
 * blocks metadata.
 *
 * This functor is used to implement a lazy initialization. It initialy
 * reference a temporary function to trigger memtrack initialization before
 * calling the actual function. Then, it references the actual function directly
 * to avoid any useless tests.
 *
 *  @return Number of registered blocks
 */
extern uint64_t (*memtrack_getallocatedblocks) ();

/** Functor to get the total RAM size allocated
 *
 * This function returns the internal summ of all the allocated memory blocks
 * which are registered.
 *
 * This functor is used to implement a lazy initialization. It initialy
 * reference a temporary function to trigger memtrack initialization before
 * calling the actual function. Then, it references the actual function directly
 * to avoid any useless tests.
 *
 *  @return Total RAM size in bytes
 */
extern uint64_t (*memtrack_getallocatedRAM) ();

/** Functor to get size of a specific memory block
 *
 * The function will search in the list for the specified pointer. If the
 * pointer is not found, it will return 0, which is discriminent as the
 * memtracker does not allow to track a zero sized block. The memtracker does
 * not allow neither to track a NULL pointer, thus NULL will return 0.
 * Otherwise, the function will return the memory block size.
 *
 * This functor is used to implement a lazy initialization. It initialy
 * reference a temporary function to trigger memtrack initialization before
 * calling the actual function. Then, it references the actual function directly
 * to avoid any useless tests.
 *
 *  @param [in] p_Ptr Allocated and tracked memory block pointer
 *
 *  @return Memory block size in bytes
 */
extern size_t(*memtrack_getblocksize) (const void *p_Ptr);

#ifdef __cplusplus
}
#endif

#endif                          /* __MEMTRACK_H__ */

