/**   \file  suite_libdebug-memdbg.c
 *   \brief  Check Libdebug memdbg functions
 *  \author  François Cerbelle (Fanfan), francois@cerbelle.net
 *
 *  \internal
 *       Created:  26/06/2024
 *      Revision:  none
 * Last modified:  2024-07-26 14:45
 *      Compiler:  gcc
 *  Organization:  Cerbelle.net
 *     Copyright:  Copyright (c) 2024, François Cerbelle
 *
 *  This source code is released for free distribution under the terms of the
 *  GNU General Public License as published by the Free Software Foundation.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "debug/memdbg.h"
#include "debug/oom.h"                                /* OOM simulation */
#include <check.h>
#include <unistd.h>                             /* pid_t getpid() */
#include <sys/stat.h>                           /* mkdir(), chdir(), mode_t */
#include <stdio.h>                              /* printf() */
#include <stdlib.h>                             /* abort() */
#include "checktools.inc"

#ifndef NDEBUG
START_TEST(memdbg_malloc_0)
{
    uint64_t l_NbBlocks;
    size_t l_RAMSize;
    void* result;

    /* LibMemTrack consistency check preparation */
    l_NbBlocks=memtrack_getallocatedblocks();
    l_RAMSize=memtrack_getallocatedRAM();
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");

    oomtest_fill(_i,2);
    /* Malloc has a non portable behavior, see man page, for 0 sized blocks */
    /* We should not support it */
    result=dbg_malloc(
               0,
               __FILE__, __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_malloc_null)
{
    uint64_t l_NbBlocks;
    size_t l_RAMSize;
    void* result;

    /* LibMemTrack consistency check preparation */
    l_NbBlocks=memtrack_getallocatedblocks();
    l_RAMSize=memtrack_getallocatedRAM();
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");

    oomtest_fill(_i,2);
    result=dbg_malloc(
               1,
               __FILE__,0,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());

    oomtest_fill(_i,2);
    result=dbg_malloc(
               1,
               "", __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());

    oomtest_fill(_i,2);
    result=dbg_malloc(
               1,
               __FILE__, __LINE__,
               "", __TIME__,__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());

    oomtest_fill(_i,2);
    result=dbg_malloc(
               1,
               __FILE__, __LINE__,
               __DATE__, "",__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());

    oomtest_fill(_i,2);
    result=dbg_malloc(
               1,
               __FILE__, __LINE__,
               __DATE__, __TIME__,""
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_free_null)
{
    uint64_t l_NbBlocks;
    size_t l_RAMSize;
    void* result=NULL;

    /* LibMemTrack consistency check preparation */
    l_NbBlocks=memtrack_getallocatedblocks();
    l_RAMSize=memtrack_getallocatedRAM();
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");

    /* should work */
    oomtest_fill(_i,2);
    dbg_free(NULL,
             __FILE__, __LINE__,
             __DATE__, __TIME__,__func__
            );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_free_invalid)
{
    uint64_t l_NbBlocks;

    /* LibMemTrack consistency check preparation */
    l_NbBlocks=memtrack_getallocatedblocks();
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");

    /* This test is supposed to trigger a SIGABRT(6) and will crash the
     * whole testsuite if not caught or not in a child process */
    forktest_only;
#ifdef GCOV
    if (oomtest_enabled())
        __gcov_dump();
#endif
    oomtest_fill(_i,2);
    dbg_free((void*)123,
             __FILE__, __LINE__,
             __DATE__, __TIME__,__func__
            );
    /* Never reached, dead code */
}
END_TEST
#endif


#ifndef NDEBUG
START_TEST(memdbg_mallocfree)
{
    uint64_t l_NbBlocks;
    size_t l_RAMSize;
    void* result;

    /* LibMemTrack consistency check preparation */
    l_NbBlocks=memtrack_getallocatedblocks();
    l_RAMSize=memtrack_getallocatedRAM();
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");

    /* Allocate */
    oomtest_fill(_i,2);
    result=dbg_malloc(
               10000,
               __FILE__, __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert((NULL!=result)||(oomtest_enabled()));
    if (NULL!=result) {
        /* Function returned the expected result */
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(l_RAMSize+10000==memtrack_getallocatedRAM());
    } else {
        /* Function failed to return expected result while in OOM mode */
        ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
        ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    }

    /* Deallocate */
    oomtest_fill(_i,2);
    dbg_free(result,
             __FILE__, __LINE__,
             __DATE__, __TIME__,__func__
            );
    oomtest_free();
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_calloc_0)
{
    uint64_t l_NbBlocks;
    size_t l_RAMSize;
    void* result;

    /* LibMemTrack consistency check preparation */
    l_NbBlocks=memtrack_getallocatedblocks();
    l_RAMSize=memtrack_getallocatedRAM();
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");

    oomtest_fill(_i,2);
    /* As well as malloc, calloc returns a pointer for a zero sized block allocation */
    /* We do not want to support this */
    result=dbg_calloc(
               0, 0,
               __FILE__, __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());

    oomtest_fill(_i,2);
    result=dbg_calloc(
               1, 0,
               __FILE__, __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());

    oomtest_fill(_i,2);
    result=dbg_calloc(
               0, 1,
               __FILE__, __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_calloc_null)
{
    uint64_t l_NbBlocks;
    size_t l_RAMSize;
    void* result;

    /* LibMemTrack consistency check preparation */
    l_NbBlocks=memtrack_getallocatedblocks();
    l_RAMSize=memtrack_getallocatedRAM();
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");

    oomtest_fill(_i,2);
    result=dbg_calloc(
               1,1,
               __FILE__,0,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());

    oomtest_fill(_i,2);
    result=dbg_calloc(
               1,1,
               "", __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());

    oomtest_fill(_i,2);
    result=dbg_calloc(
               1,1,
               __FILE__, __LINE__,
               "", __TIME__,__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());

    oomtest_fill(_i,2);
    result=dbg_calloc(
               1,1,
               __FILE__, __LINE__,
               __DATE__, "",__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());

    oomtest_fill(_i,2);
    result=dbg_calloc(
               1,1,
               __FILE__, __LINE__,
               __DATE__, __TIME__,""
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_callocfree)
{
    uint64_t l_NbBlocks;
    size_t l_RAMSize;
    void* result;

    /* LibMemTrack consistency check preparation */
    l_NbBlocks=memtrack_getallocatedblocks();
    l_RAMSize=memtrack_getallocatedRAM();
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");

    /* Allocate */
    oomtest_fill(_i,2);
    result=dbg_calloc(
               100,100,
               __FILE__, __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert((NULL!=result)||(oomtest_enabled()));
    if (NULL!=result) {
        /* Function returned the expected result */
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(l_RAMSize+10000==memtrack_getallocatedRAM());
    } else {
        /* Function failed to return expected result while in OOM mode */
        ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
        ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    }

    /* Deallocate */
    oomtest_fill(_i,2);
    dbg_free(result,
             __FILE__, __LINE__,
             __DATE__, __TIME__,__func__
            );
    oomtest_free();
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
}
END_TEST
#endif

/* WIP */
#ifndef NDEBUG
START_TEST(memdbg_realloc_null0)
{
    uint64_t l_NbBlocks;
    size_t l_RAMSize;
    void* l_Ptr;
    void* result;

    /* LibMemTrack consistency check preparation */
    l_NbBlocks=memtrack_getallocatedblocks();
    l_RAMSize=memtrack_getallocatedRAM();
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");

    /* Null test */
    /* Should behave as malloc in this case and allocate a valid initial block */
    l_Ptr=NULL;
    oomtest_fill(_i,2);
    result=dbg_realloc(
               l_Ptr,
               10,
               __FILE__,__LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert((NULL!=result)||(oomtest_enabled()));
    if (NULL!=result) {
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(l_RAMSize+10==memtrack_getallocatedRAM());
        ck_assert_msg(l_NbBlocks+1==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");
    } else {
        ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
        ck_assert(l_RAMSize==memtrack_getallocatedRAM());
        ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");
        /* Repair for next test */
        result=dbg_realloc(
                   l_Ptr,
                   10,
                   __FILE__,__LINE__,
                   __DATE__, __TIME__,__func__
               );
    }

    /* Non portable behavior which behaves as free */
    l_Ptr=result;
    oomtest_fill(_i,2);
    result=dbg_realloc(
               l_Ptr,
               0,
               __FILE__, __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert(l_Ptr==result);
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_realloc_empty)
{
    uint64_t l_NbBlocks;
    size_t l_RAMSize;
    void* l_Ptr;
    void* result;

    /* LibMemTrack consistency check preparation */
    l_NbBlocks=memtrack_getallocatedblocks();
    l_RAMSize=memtrack_getallocatedRAM();
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");

    /* Create an initial valid block */
    l_Ptr=NULL;
    oomtest_fill(_i,2);
    result=dbg_realloc(
               l_Ptr,
               10,
               __FILE__,__LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert((NULL!=result)||(oomtest_enabled()));
    if (NULL!=result) {
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(l_RAMSize+10==memtrack_getallocatedRAM());
    } else {
        ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
        ck_assert(l_RAMSize==memtrack_getallocatedRAM());
        /* Repair for next test */
        result=dbg_realloc(
                   l_Ptr,
                   10,
                   __FILE__,__LINE__,
                   __DATE__, __TIME__,__func__
               );
    }

    /* Line number 0 */
    l_Ptr=result;
    oomtest_fill(_i,2);
    result=dbg_realloc(
               l_Ptr,
               1,
               __FILE__,0,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize+10==memtrack_getallocatedRAM());

    /* Empty filename */
    oomtest_fill(_i,2);
    result=dbg_realloc(
               l_Ptr,
               1,
               "", __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize+10==memtrack_getallocatedRAM());

    /* Empty Compilation date */
    oomtest_fill(_i,2);
    result=dbg_realloc(
               l_Ptr,
               1,
               __FILE__, __LINE__,
               "", __TIME__,__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize+10==memtrack_getallocatedRAM());

    /* Empty Compilation time */
    oomtest_fill(_i,2);
    result=dbg_realloc(
               l_Ptr,
               1,
               __FILE__, __LINE__,
               __DATE__, "",__func__
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize+10==memtrack_getallocatedRAM());

    /* Empty function name */
    oomtest_fill(_i,2);
    result=dbg_realloc(
               l_Ptr,
               1,
               __FILE__, __LINE__,
               __DATE__, __TIME__,""
           );
    oomtest_free();
    ck_assert(NULL==result);
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize+10==memtrack_getallocatedRAM());

    dbg_free(
        l_Ptr,
        __FILE__, __LINE__,
        __DATE__, __TIME__,__func__
    );
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_realloc_invalid)
{
    uint64_t l_NbBlocks;

    /* LibMemTrack consistency check preparation */
    l_NbBlocks=memtrack_getallocatedblocks();
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");

    /* This test is supposed to trigger a SIGABRT(6) and will crash the
     * whole testsuite if not caught or not in a child process */
    forktest_only;

#ifdef GCOV
    if (oomtest_enabled())
        __gcov_dump();
#endif
    oomtest_fill(_i,2);
    /* Try to realloc a non null, not tracked block */
    /* Abort/SIG6 expected */
    dbg_realloc(
        (void*)999,
        10,
        __FILE__, __LINE__, __DATE__, __TIME__, __func__
    );
    /* Never reached, dead code */
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_reallocfree)
{
    uint64_t l_NbBlocks;
    size_t l_RAMSize;
    void* l_Ptr;
    void* result;

    /* LibMemTrack consistency check preparation */
    l_NbBlocks=memtrack_getallocatedblocks();
    l_RAMSize=memtrack_getallocatedRAM();
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");

    /* Create an initial valid block */
    l_Ptr=NULL;
    result=dbg_malloc(
               10,
               __FILE__,__LINE__,
               __DATE__, __TIME__,__func__
           );
    ck_assert(NULL!=result);
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize+10==memtrack_getallocatedRAM());

    /* Standard realloc to increase the block size */
    l_Ptr=result;
    oomtest_fill(_i,2);
    result = dbg_realloc(
                 l_Ptr,
                 10000,
                 __FILE__, __LINE__, __DATE__, __TIME__, __func__
             );
    oomtest_free();
    ck_assert((NULL!=result)||(oomtest_enabled()));
    if (NULL!=result) {
        /* Function returned the expected result */
    } else {
        /* Function failed to return expected result while in OOM mode */
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(l_RAMSize+10==memtrack_getallocatedRAM());
        /* Repair for next test */
        result = dbg_realloc(
                     l_Ptr,
                     10000,
                     __FILE__, __LINE__, __DATE__, __TIME__, __func__
                 );
        ck_assert(NULL!=result);
    }
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize+10000==memtrack_getallocatedRAM());

    l_Ptr=result;
    /* Standard realloc to reduce the block size */
    oomtest_fill(_i,2);
    result = dbg_realloc(
                 l_Ptr,
                 1000,
                 __FILE__, __LINE__, __DATE__, __TIME__, __func__
             );
    oomtest_free();
    ck_assert((NULL!=result)||(oomtest_enabled()));
    if (NULL!=result) {
        /* Function returned the expected result */
    } else {
        /* Function failed to return expected result while in OOM mode */
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(l_RAMSize+10000==memtrack_getallocatedRAM());
        /* Repair for next test */
        result = dbg_realloc(
                     l_Ptr,
                     1000,
                     __FILE__, __LINE__, __DATE__, __TIME__, __func__
                 );
    }
    ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize+1000==memtrack_getallocatedRAM());

    /* OOM expected, no realloc */
    l_Ptr=result;
    if (oomtest_enabled()) {
        oomtest_fill(_i,2);
        result= dbg_realloc(
                    l_Ptr,
                    (RAMLIMIT_SOFT+1),
                    __FILE__, __LINE__, __DATE__, __TIME__, __func__
                );
        oomtest_free();

        ck_assert(NULL==result);
        ck_assert(l_NbBlocks+1==memtrack_getallocatedblocks());
        ck_assert(l_RAMSize+1000==memtrack_getallocatedRAM());
    }
    /* Optional cleanup when forked, but cleaner and dbg_free already tested */
    dbg_free(
        l_Ptr,
        __FILE__, __LINE__,
        __DATE__, __TIME__,__func__
    );
    ck_assert(l_NbBlocks==memtrack_getallocatedblocks());
    ck_assert(l_RAMSize==memtrack_getallocatedRAM());
    ck_assert_msg(l_NbBlocks==memtrack_dumpblocks(),"Internal NbBlocks counter and dumpblocks mismatch");
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_strdup_invalid)
{
    /* This test is supposed to trigger a SIGSEV(11) and will crash the
     * whole testsuite if not caught or not in a child process */
    forktest_only;

    oomtest_fill(_i,2);
    /* SIGSEV expected */
    dbg_strdup(
        NULL,
        __FILE__, __LINE__,
        __DATE__, __TIME__,__func__
    );
    /* Never reached, dead code */
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_strdup_nullempty)
{
    char* result;

    result = NULL;
    oomtest_fill(_i,2);
    result=dbg_strdup(
               "",
               __FILE__, __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert((NULL!=result)||(oomtest_enabled()));
    dbg_free(result,
             __FILE__, __LINE__,
             __DATE__, __TIME__,__func__
            );

    result = NULL;
    oomtest_fill(_i,2);
    result=dbg_strdup(
               "abc",
               NULL, __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert((NULL==result)||(oomtest_enabled()));

    result = NULL;
    oomtest_fill(_i,2);
    result=dbg_strdup(
               "abc",
               __FILE__, 0,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert((NULL==result)||(oomtest_enabled()));

    result = NULL;
    oomtest_fill(_i,2);
    result=dbg_strdup(
               "abc",
               __FILE__, __LINE__,
               NULL, __TIME__,__func__
           );
    oomtest_free();
    ck_assert((NULL==result)||(oomtest_enabled()));

    result = NULL;
    oomtest_fill(_i,2);
    result=dbg_strdup(
               "abc",
               __FILE__, __LINE__,
               __DATE__, NULL,__func__
           );
    oomtest_free();
    ck_assert((NULL==result)||(oomtest_enabled()));

    result = NULL;
    oomtest_fill(_i,2);
    result=dbg_strdup(
               "abc",
               __FILE__, __LINE__,
               __DATE__, __TIME__,NULL
           );
    oomtest_free();
    ck_assert((NULL==result)||(oomtest_enabled()));

    result = NULL;
    oomtest_fill(_i,2);
    result=dbg_strdup(
               "abc",
               "", __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert((NULL==result)||(oomtest_enabled()));

    result = NULL;
    oomtest_fill(_i,2);
    result=dbg_strdup(
               "abc",
               __FILE__, __LINE__,
               "", __TIME__,__func__
           );
    oomtest_free();
    ck_assert((NULL==result)||(oomtest_enabled()));

    result = NULL;
    oomtest_fill(_i,2);
    result=dbg_strdup(
               "abc",
               __FILE__, __LINE__,
               __DATE__, "",__func__
           );
    oomtest_free();
    ck_assert((NULL==result)||(oomtest_enabled()));

    result = NULL;
    oomtest_fill(_i,2);
    result=dbg_strdup(
               "abc",
               __FILE__, __LINE__,
               __DATE__, __TIME__,""
           );
    oomtest_free();
    ck_assert((NULL==result)||(oomtest_enabled()));
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_strdup_value)
{
    char* result;
    char *szLongString =
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "%d"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        ;
    oomtest_fill(_i,2);
    result=dbg_strdup(
               szLongString,
               __FILE__, __LINE__,
               __DATE__, __TIME__,__func__
           );
    oomtest_free();
    ck_assert((NULL!=result)||(oomtest_enabled()));
    if (NULL!=result) {
        ck_assert(0==strcmp(szLongString,result));
        dbg_free(result,
                 __FILE__, __LINE__,
                 __DATE__, __TIME__,__func__
                );
    }
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_asprintf_invalid)
{
    forktest_only;
#ifdef GCOV
    if (oomtest_enabled())
        __gcov_dump();
#endif
    oomtest_fill(_i,2);
    /* SIGABRT6 expected */
    dbg_asprintf(NULL,
                 "abc%ddef",
                 __FILE__, __LINE__,
                 __DATE__, __TIME__,__func__,
                 123
                );
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_asprintf_nullempty)
{
    int result;
    char* l_Ptr;

    oomtest_fill(_i,2);
    result=dbg_asprintf(&l_Ptr,
                        NULL,
                        __FILE__, __LINE__,
                        __DATE__, __TIME__,__func__
                       );
    oomtest_free();
    ck_assert((-1==result)||((oomtest_enabled())&&(-2==result)));

    oomtest_fill(_i,2);
    result=dbg_asprintf(&l_Ptr,
                        "abcdef",
                        NULL, __LINE__,
                        __DATE__, __TIME__,__func__
                       );
    oomtest_free();
    ck_assert((-2==result)||((oomtest_enabled())&&(-1==result)));

    oomtest_fill(_i,2);
    result=dbg_asprintf(&l_Ptr,
                        "abcdef",
                        __FILE__, 0,
                        __DATE__, __TIME__,__func__
                       );
    oomtest_free();
    ck_assert((-2==result)||((oomtest_enabled())&&(-1==result)));

    oomtest_fill(_i,2);
    result=dbg_asprintf(&l_Ptr,
                        "abcdef",
                        __FILE__, __LINE__,
                        NULL, __TIME__,__func__
                       );
    oomtest_free();
    ck_assert((-2==result)||((oomtest_enabled())&&(-1==result)));

    oomtest_fill(_i,2);
    result=dbg_asprintf(&l_Ptr,
                        "abcdef",
                        __FILE__, __LINE__,
                        __DATE__, NULL,__func__
                       );
    oomtest_free();
    ck_assert((-2==result)||((oomtest_enabled())&&(-1==result)));

    oomtest_fill(_i,2);
    result=dbg_asprintf(&l_Ptr,
                        "abcdef",
                        __FILE__, __LINE__,
                        __DATE__, __TIME__,NULL
                       );
    oomtest_free();
    ck_assert((-2==result)||((oomtest_enabled())&&(-1==result)));

    oomtest_fill(_i,2);
    result=dbg_asprintf(&l_Ptr,
                        "",
                        __FILE__, __LINE__,
                        __DATE__, __TIME__,__func__
                       );
    oomtest_free();
    ck_assert((0==result)||(oomtest_enabled()));
    if (0==result) {
        ck_assert(NULL!=l_Ptr);
        dbg_free(l_Ptr,
                 __FILE__, __LINE__,
                 __DATE__, __TIME__,__func__
                );
    }

    oomtest_fill(_i,2);
    result=dbg_asprintf(&l_Ptr,
                        "abcdef",
                        "", __LINE__,
                        __DATE__, __TIME__,__func__
                       );
    oomtest_free();
    ck_assert((-2==result)||((oomtest_enabled())&&(-1==result)));

    oomtest_fill(_i,2);
    result=dbg_asprintf(&l_Ptr,
                        "abcdef",
                        __FILE__, 0,
                        __DATE__, __TIME__,__func__
                       );
    oomtest_free();
    ck_assert((-2==result)||((oomtest_enabled())&&(-1==result)));

    oomtest_fill(_i,2);
    result=dbg_asprintf(&l_Ptr,
                        "abcdef",
                        __FILE__, __LINE__,
                        "", __TIME__,__func__
                       );
    oomtest_free();
    ck_assert((-2==result)||((oomtest_enabled())&&(-1==result)));

    oomtest_fill(_i,2);
    result=dbg_asprintf(&l_Ptr,
                        "abcdef",
                        __FILE__, __LINE__,
                        __DATE__, "",__func__
                       );
    oomtest_free();
    ck_assert((-2==result)||((oomtest_enabled())&&(-1==result)));

    oomtest_fill(_i,2);
    result=dbg_asprintf(&l_Ptr,
                        "abcdef",
                        __FILE__, __LINE__,
                        __DATE__, __TIME__,""
                       );
    oomtest_free();
    ck_assert((-2==result)||((oomtest_enabled())&&(-1==result)));
}
END_TEST
#endif

#ifndef NDEBUG
START_TEST(memdbg_asprintf_value)
{
    /* Only test with parameter, not without or with parameter type error */
    int result;
    char* l_Ptr;
    char *szLongString =
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "%d"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        ;
    oomtest_fill(_i,2);
    result=dbg_asprintf(&l_Ptr,
                        szLongString,
                        __FILE__, __LINE__,
                        __DATE__, __TIME__,__func__,
                        123
                       );
    oomtest_free();
    ck_assert((-1!=result)||(oomtest_enabled()));
    if (-1!=result) {
        char *szExpected =
            "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "123"
            "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            ;
        /* Test succeeded */
        ck_assert((size_t)result==strlen(szExpected));
        ck_assert((NULL!=l_Ptr)&&(0==strcmp(szExpected,l_Ptr)));
        dbg_free(l_Ptr,
                 __FILE__, __LINE__,
                 __DATE__, __TIME__,__func__
                );
    }
}
END_TEST
#endif

/************************************************************************/
void memdbg_checked_uninitialized_setup()
{
    signals_catch();
    forktest_gprofdir();
}
void memdbg_checked_withdata_setup()
{
    signals_catch();
    forktest_gprofdir();
}
void memdbg_checked_common_teardown()
{
    signals_release();
#ifndef NDEBUG
    ck_assert(0==memtrack_getallocatedblocks());
#endif /* NDEBUG */
}
/************************************************************************/
void memdbg_unchecked_common_setup()
{
    forktest_init();
}
void memdbg_unchecked_common_teardown()
{
}
void memdbg_unchecked_oom_setup()
{
    oomtest_enable(RAMLIMIT_SOFT);
    memdbg_unchecked_common_setup();
}
void memdbg_unchecked_oom_teardown()
{
    oomtest_disable();
    memdbg_unchecked_common_teardown();
}
/************************************************************************/
Suite* libdebug_memdbg_suite()
{
    Suite *s;
    TCase *tc;

    s = suite_create("LibDebug-MemDbg");
    tc = tcase_create("Uninitialized");
    tcase_add_checked_fixture(tc, memdbg_checked_uninitialized_setup, memdbg_checked_common_teardown);
    tcase_add_unchecked_fixture(tc, memdbg_unchecked_common_setup, memdbg_unchecked_common_teardown);
    suite_add_tcase(s, tc);
#ifndef NDEBUG
    tcase_add_test(tc, memdbg_malloc_0);
    tcase_add_test(tc, memdbg_malloc_null);
    tcase_add_test(tc, memdbg_free_null);
    tcase_add_test_raise_signal(tc, memdbg_free_invalid, 6);
    tcase_add_test(tc, memdbg_mallocfree);
    tcase_add_test(tc, memdbg_calloc_0);
    tcase_add_test(tc, memdbg_calloc_null);
    tcase_add_test(tc, memdbg_callocfree);
    tcase_add_test(tc, memdbg_realloc_null0);
    tcase_add_test(tc, memdbg_realloc_empty);
    tcase_add_test_raise_signal(tc, memdbg_realloc_invalid, 6);
    tcase_add_test(tc, memdbg_reallocfree);
    tcase_add_test_raise_signal(tc, memdbg_strdup_invalid, 11);
    tcase_add_test(tc, memdbg_strdup_nullempty);
    tcase_add_test(tc, memdbg_strdup_value);
    tcase_add_test_raise_signal(tc, memdbg_asprintf_invalid, 6);
    tcase_add_test(tc, memdbg_asprintf_nullempty);
    tcase_add_test(tc, memdbg_asprintf_value);
#endif

    tc = tcase_create("WithData");
    tcase_add_checked_fixture(tc, memdbg_checked_withdata_setup, memdbg_checked_common_teardown);
    tcase_add_unchecked_fixture(tc, memdbg_unchecked_common_setup, memdbg_unchecked_common_teardown);
    suite_add_tcase(s, tc);
#ifndef NDEBUG
    tcase_add_test(tc, memdbg_malloc_0);
    tcase_add_test(tc, memdbg_malloc_null);
    tcase_add_test(tc, memdbg_free_null);
    tcase_add_test_raise_signal(tc, memdbg_free_invalid, 6);
    tcase_add_test(tc, memdbg_mallocfree);
    tcase_add_test(tc, memdbg_calloc_0);
    tcase_add_test(tc, memdbg_calloc_null);
    tcase_add_test(tc, memdbg_callocfree);
    tcase_add_test(tc, memdbg_realloc_null0);
    tcase_add_test(tc, memdbg_realloc_empty);
    tcase_add_test_raise_signal(tc, memdbg_realloc_invalid, 6);
    tcase_add_test(tc, memdbg_reallocfree);
    tcase_add_test_raise_signal(tc, memdbg_strdup_invalid, 11);
    tcase_add_test(tc, memdbg_strdup_nullempty);
    tcase_add_test(tc, memdbg_strdup_value);
    tcase_add_test_raise_signal(tc, memdbg_asprintf_invalid, 6);
    tcase_add_test(tc, memdbg_asprintf_nullempty);
    tcase_add_test(tc, memdbg_asprintf_value);
#endif

    tc = tcase_create("Uninitialized-OOM");
    tcase_set_tags(tc,"oom");
    tcase_add_checked_fixture(tc, memdbg_checked_uninitialized_setup, memdbg_checked_common_teardown);
    tcase_add_unchecked_fixture(tc, memdbg_unchecked_oom_setup, memdbg_unchecked_oom_teardown);
    suite_add_tcase(s, tc);
#ifndef NDEBUG
    tcase_add_loop_test(tc, memdbg_malloc_0, 0, 10);
    tcase_add_loop_test(tc, memdbg_malloc_null, 0, 10);
    tcase_add_loop_test(tc, memdbg_free_null, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memdbg_free_invalid, 6, 0, 10);
    tcase_add_loop_test(tc, memdbg_mallocfree, 0, 10);
    tcase_add_loop_test(tc, memdbg_calloc_0, 0, 10);
    tcase_add_loop_test(tc, memdbg_calloc_null, 0, 10);
    tcase_add_loop_test(tc, memdbg_callocfree, 0, 10);
    tcase_add_loop_test(tc, memdbg_realloc_null0, 0, 10);
    tcase_add_loop_test(tc, memdbg_realloc_empty, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memdbg_realloc_invalid, 6, 0, 10);
    tcase_add_loop_test(tc, memdbg_reallocfree, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memdbg_strdup_invalid, 11, 0, 10);
    tcase_add_loop_test(tc, memdbg_strdup_nullempty, 0, 10);
    tcase_add_loop_test(tc, memdbg_strdup_value, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memdbg_asprintf_invalid, 6, 0, 10);
    tcase_add_loop_test(tc, memdbg_asprintf_nullempty, 0, 10);
    tcase_add_loop_test(tc, memdbg_asprintf_value, 0, 10);
#endif

    tc = tcase_create("WithData-OOM");
    tcase_set_tags(tc,"oom");
    tcase_add_checked_fixture(tc, memdbg_checked_withdata_setup, memdbg_checked_common_teardown);
    tcase_add_unchecked_fixture(tc, memdbg_unchecked_oom_setup, memdbg_unchecked_oom_teardown);
    suite_add_tcase(s, tc);
#ifndef NDEBUG
    tcase_add_loop_test(tc, memdbg_malloc_0, 0, 10);
    tcase_add_loop_test(tc, memdbg_malloc_null, 0, 10);
    tcase_add_loop_test(tc, memdbg_free_null, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memdbg_free_invalid, 6, 0, 10);
    tcase_add_loop_test(tc, memdbg_mallocfree, 0, 10);
    tcase_add_loop_test(tc, memdbg_calloc_0, 0, 10);
    tcase_add_loop_test(tc, memdbg_calloc_null, 0, 10);
    tcase_add_loop_test(tc, memdbg_callocfree, 0, 10);
    tcase_add_loop_test(tc, memdbg_realloc_null0, 0, 10);
    tcase_add_loop_test(tc, memdbg_realloc_empty, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memdbg_realloc_invalid, 6, 0, 10);
    tcase_add_loop_test(tc, memdbg_reallocfree, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memdbg_strdup_invalid, 11, 0, 10);
    tcase_add_loop_test(tc, memdbg_strdup_nullempty, 0, 10);
    tcase_add_loop_test(tc, memdbg_strdup_value, 0, 10);
    tcase_add_loop_test_raise_signal(tc, memdbg_asprintf_invalid, 6, 0, 10);
    tcase_add_loop_test(tc, memdbg_asprintf_nullempty, 0, 10);
    tcase_add_loop_test(tc, memdbg_asprintf_value, 0, 10);
#endif

    return s;
}

